#include "field.h"
#include "window.h"
#include "keyboard.h"
#include "isam.h"
#include "cbtree.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#if ! defined(OK)
   #define OK I_OK
#endif
#if ! defined(ERROR)
   #define ERROR I_ERROR
#endif

/* ------------------------------------------------------------ */
/*  edit a record in a database                                 */
/*    The record is displayed in the window with one field      */
/*    per line.  After editing, the fields are read and the     */
/*    current record is replaced by the new values unless the   */
/*    field input is terminated by the Esc key.                 */

/*  Returns:   */
/*     -1 if a window error or out of memory error occured */
/*     -2 if there was a database error in imodrec */
/*    the key that terminated the f_batch if no error */
/*        if the key is _ESC, the database is not updated */

int db_edit_record(Window wd, Db_Obj *db, Index_Obj *index)
{
   int field_count;
   int record_size;
   int max_name = 0;
   int max_width = 0;
   int error = 0;
   int type, ii, sign;
   int len, idx, *iptr;
   int status;
   unsigned key;
   long ll, *lptr;
   double dd, *dptr;
   char *newbuf, *buf;
   char *buffer;
   char **ptrs, **ptr;
   char **names;
   char *p;
   char prompt[80], template[80];
   Field fld;

   if (f_getfield(wd,0) != NULL) return -1;  /* window already has fields */
   field_count = igetfldcount(db);        /* Number of fields in database */
   if (field_count < 0) return -1;        /* Invalid database */
   /* if the window does not have one row per field, expand it */
   if (w_info(wd, IROWS) < field_count) {
      if (w_size(wd, w_info(wd, ICOLS), field_count) != 0) return -1;
      }
   /* read the current record using the supplied index */
   if (igetreclen(db, index, &record_size) != OK) return -1;
   buffer = malloc(record_size);
   if (buffer == NULL) return -1;                /* No space for buffer */
   ptrs = malloc((field_count+1) * sizeof(char *));
   if (ptrs == NULL) {free(buffer); return -1;}  /* no space for pointers */
   if (igetrec(db, index, ptrs, buffer, record_size) != OK) {
      free(buffer);
      free(ptrs);
      return -1;
      }

   /* find the length of the longest field name to determine the width */
   /* of the field prompts.  Also find the width of the longest field  */
   /* to expand the window so that the entire field can be edited.     */
   names = igetfldnames(db);
   ptr = ptrs;
   for (idx = 0; idx < field_count; idx++) {
      if ((p = strchr(*names, ',')) != NULL) len = p - *names;
      else len = strlen(*names);
      if (len > max_name) max_name = len;
      if (p != NULL) type = toupper(*(p+1)); else type = 'S';
      switch (type) {   /* type of field */
         case 'C':
         case 'S': len = strlen(*ptr); break;
         case 'I':
         case 'U': len = 6; break;
         case 'L': len = 11; break;
         case 'F': len = 11; break;
         case 'D': len = 21; break;
         case 'B':      /* editing of binary fields is not supported */
         default:       /* unknown type of field */
            free(buffer); free(ptrs); return -1;
         }
      if (len > max_width) max_width = len;
      names++;
      ptr++;
      }
   if (max_width >= sizeof(template) || max_name >= sizeof(prompt)-2) {
      free(buffer); free(ptrs); return -1;  /* arrays are too small */
      }

   /* expand the window to allow editing of the longest current field */
   if ((max_name + max_width + 3) > w_info(wd, ICOLS)) {
      if (w_size(wd, max_name+max_width+3, w_info(wd, IROWS)) != 0) {
         free(buffer); free(ptrs); return -1;
         }
      }
   else max_width = w_info(wd, ICOLS) - max_name - 3;  /* use full window */
   newbuf = malloc((max_width+1)*field_count);  /* buffer for result */
   if (newbuf == NULL) {free(buffer); free(ptrs); return -1;}

   /* create a data entry field for each database field */
   w_clear(wd);
   names = igetfldnames(db);
   ptr = ptrs;
   for (idx = 0; idx < field_count; idx++) {
      if ((p = strchr(*names, ',')) != NULL) {
         len = p - *names;
         type = toupper(*(p+1));
         }
      else {len = strlen(*names); type = 'S';}
      if (len < max_name) memset(prompt, ' ', max_name-len);/* leading blanks */
      memcpy(&prompt[max_name-len], *names, len);  /* field name */
      memcpy(&prompt[max_name], ": ", 3);          /* separator */
      switch (type) {   /* type of field */
         case 'C':
         case 'S': len = max_width; break;
         case 'I':
         case 'U': len = 6; break;
         case 'L': len = 11; break;
         case 'F': len = 11; break;
         case 'D': len = 21; break;
         }
      memset(template, '_', len);       /* field template */
      template[len] = '\0';
      fld = f_create(prompt, template);      /* create the field */
      if (fld == NULL) error = 1;   /* unable to create a field */
      else {
         /* initialize field with current value */
         if (p != NULL) type = toupper(*(p+1)); else type = 'S';
         switch (type) {   /* type of field */
            case 'C':
            case 'S': f_setbuffer(fld, *ptr); break;
            case 'I':
               ii = * ((int *) (*ptr));
               sign = ii < 0;
               f_setint(fld, ii, ENABLE, sign, ENABLE);
               break;
            case 'U':
               ll = *((unsigned *) (*ptr));
               f_setlong(fld, ll, ENABLE, DISABLE, ENABLE);
               break;
            case 'L':
               ll = *((unsigned *) (*ptr));
               sign = ll < 0;
               f_setlong(fld, ll, ENABLE, sign, ENABLE);
               break;
            case 'F':
            case 'D':
               if (type == 'F') dd = *((float *) (*ptr));
               else dd = *((double *) (*ptr));
               sign = dd < 0;
               f_setfloat(fld, dd, ENABLE, sign);
               break;
            }
         if (f_append(wd, 0, idx, fld) != 0) error=1;  /* add to the window */
         names++;                      /* move to next field */
         ptr++;
         }  /* field create is ok */
      }     /* for idx */

   if (error == 0) {
      /* edit the information */
      key = f_batch(wd);
      status = key;

      /* read the new fields */
      if (key != _ESC) {
         ptr = ptrs;
         names = igetfldnames(db);
         buf = newbuf;
         for (idx = 0; idx < field_count; idx++) {
            fld = f_getfield(wd, idx);       /* get the field */
            p = strchr(*names, ',');         /* find its type */
            if (p != NULL) type = toupper(*(p+1)); else type = 'S';
            switch (type) {
               case 'C':
               case 'S':
                  f_getstring(fld, buf);
                  *ptr = buf;
                  buf += strlen(buf) + 1;
                  break;
               case 'I':
                  iptr = (int *) buf;
                  *iptr = f_getint(fld);
                  *ptr = buf;
                  buf += sizeof(int);
                  break;
               case 'U':
                  ll = f_getlong(fld);
                  *((unsigned *) buf) = (unsigned) ll;
                  *ptr = buf;
                  buf += sizeof(unsigned);
                  break;
               case 'L':
                  lptr = (long *) buf;
                  *lptr = f_getlong(fld);
                  *ptr = buf;
                  buf += sizeof(long);
                  break;
               case 'F':
                  dd = f_getfloat(fld);
                  *((float *) buf) = (float) dd;
                  *ptr = buf;
                  buf += sizeof(float);
                  break;
               case 'D':
                  dptr = (double *) buf;
                  *dptr = f_getfloat(fld);
                  *ptr = buf;
                  buf += sizeof(double);
                  break;
               }
            names++;
            ptr++;
            }
         if (imodrec(db, index, ptrs) == ERROR) status = -2;
         ifindrec(db, index, ptrs);  /* Make the modified record current */
         }
      }

   /* delete the fields */
   for (idx = 0; idx < field_count; idx++) {
      fld = f_getfield(wd, 0);
      if (fld != NULL) {
         f_delete(wd, fld);
         f_free(fld);
         }
      }
   /* release the memory */
   free(newbuf);
   free(buffer);
   free(ptrs);

   return status;
   }

